'''
Python tools for creating and parsing AXI communications.
'''

import logging
from collections import namedtuple

logger = logging.getLogger(__name__)

# `Comm` objects are registered here by the name of the module they are
# responsible for communicating with.
module_register = {}

# Response code for AXI.
OKAY = 0
EXOKAY = 1
SLVERR = 2
DECERR = 3


# Used to define whether AXI command are reading or writing.
READ_TYPE = 'READ'
WRITE_TYPE = 'WRITE'


AxiResponse = namedtuple('AxiResponse', ['length', 'data', 'resp'])


class AxiResponseException(Exception):
    pass


class Future:

    UNRESOLVED = 'unresolved'
    OKAY = 'okay'
    ERROR = 'error'

    def __init__(self):
        self.result_value = None
        self.exception_value = None
        self.state = self.UNRESOLVED

    def set_result(self, result):
        assert self.state == self.UNRESOLVED
        self.result_value = result
        self.state = self.OKAY

    def set_exception(self, exception):
        assert self.state == self.UNRESOLVED
        self.exception_value = exception
        self.state = self.ERROR

    def result(self):
        if self.state == self.ERROR:
            raise self.exception_value
        elif self.state == self.UNRESOLVED:
            raise Exception('Future is unresolved')
        else:
            return self.result_value


class Command(object):

    def __init__(self, description):
        self.description = description
        self.future = Future()

    def get_axi_commands(self):
        raise Exception('Unimplemented')

    def resolve_future(self, e, result):
        if e is not None:
            self.future.set_exception(e)
        else:
            if e is not None:
                self.future.set_exception(e)
            else:
                self.future.set_result(result)

    def process_responses(self, read_responses, write_responses, resolve_future=True):
        e, r = None, None
        if resolve_future:
            self.resolve_future(e, r)
        return e, r


class AxiCommand(Command):
    '''
    Defines a series AXI4Lite master-to-slave commands.
    '''

    def __init__(self, start_address, length, readorwrite, data=None,
                 constant_address=False, description=None):
        '''
        `start_address`: The address on which the first AXI command operates.
        `constant_address`: If this is `True` we keep operating on the same
             address, otherwise we increment.
        `length`: The number of commands.
        `readorwrite`: Can be either `READ_TYPE` or `WRITE_TYPE`.
        'data': A list of integers to send (if it is a write command).
        `description`: An optional description for debugging purposes.
        '''
        max_address = pow(2, 32)-1
        self.start_address = start_address
        self.length = length
        self.readorwrite = readorwrite
        self.constant_address = constant_address
        assert readorwrite in (READ_TYPE, WRITE_TYPE)
        self.data = data
        if readorwrite == READ_TYPE:
            assert self.data is None
        else:
            assert len(self.data) == length
        assert start_address <= max_address
        if not constant_address:
            assert start_address + length-1 <= max_address
        assert description
        super().__init__(description)

    def get_axi_commands(self):
        return [self]

    def process_responses(self, read_responses, write_responses,
                          resolve_future=True):
        """
        Arguments:
          `read_responses`: An iterable of responses to read requests.  The
            iterable must begin with the requests to this command, which will
            be removed during processing.
          `write_responses`: An iterable of responses to write requests.  The
            iterable must begin with the requests to this command, which will
            be removed during processing.
          `resolve_future`: Whether the commands result future should be
            resolved.  This should be False when this command is getting
            wrapped by a higher level command.
        """
        relevant_responses = {
            READ_TYPE: read_responses,
            WRITE_TYPE: write_responses,
            }[self.readorwrite]
        data = []
        total_response_length = 0
        e = None
        while total_response_length < self.length:
            if not relevant_responses:
                e = Exception('Ran out of responses')
                break
            else:
                response = relevant_responses.pop(0)
                total_response_length += response.length
                data += response.data
                if total_response_length > self.length:
                    e = Exception(
                        'Response lengths not matching command lengths')
                elif response.resp != OKAY:
                    e = AxiResponseException(
                        'Bad response in "{}"'.format(self.description))
        # Trim data down to right size.
        # Do this so that incorrect size does not trigger errors that hide
        # the real problem.
        data = data[:self.length]
        assert len(data) == self.length
        if resolve_future:
            self.resolve_future(e, data)
        return (e, data)


class SetUnsignedsCommand(AxiCommand):

    def __init__(self, values, address, description=None,
                 constant_address=False):
        for value in values:
            assert value < pow(2, 32)
        super().__init__(
            start_address=address,
            length=len(values),
            readorwrite=WRITE_TYPE,
            data=values,
            constant_address=constant_address,
            description=description,
        )


class SetUnsignedCommand(SetUnsignedsCommand):

    def __init__(self, value, address, description=None):
        super().__init__(
            address=address,
            values=[value],
            constant_address=True,
            description=description,
        )


class SetSignedsCommand(SetUnsignedsCommand):

    def __init__(self, values, address, description=None,
                 constant_address=False):
        offset = pow(2, 32)
        unsigneds = [v+offset if v < 0 else v for v in values]
        super().__init__(
            values=unsigneds,
            address=address,
            constant_address=constant_address,
            description=description,
        )


class SetSignedCommand(SetSignedsCommand):

    def __init__(self, value, address, description=None):
        super().__init__(
            address=address,
            values=[value],
            constant_address=True,
            description=description,
        )


class GetUnsignedsCommand(AxiCommand):

    def __init__(self, address, length=1, constant_address=False,
                 description=None):
        super().__init__(
            start_address=address,
            length=length,
            readorwrite=READ_TYPE,
            constant_address=constant_address,
            description=description,
        )


class GetUnsignedCommand(AxiCommand):

    def __init__(self, address, description=None):
        super().__init__(
            start_address=address,
            length=1,
            readorwrite=READ_TYPE,
            constant_address=True,
            description=description,
        )

    def process_responses(self, read_responses, write_responses,
                          resolve_future=True):
        e, result = super().process_responses(read_responses, write_responses,
                                              resolve_future=False)
        assert len(result) == 1
        if resolve_future:
            self.resolve_future(e, result[0])
        return e, result[0]


class TriggerCommand(AxiCommand):

    def __init__(self, address, description):
        super().__init__(
            start_address=address,
            data=[0],
            length=1,
            readorwrite=WRITE_TYPE,
            constant_address=True,
            description=description,
            )


class GetBooleanCommand(AxiCommand):

    def __init__(self, address, description=None):
        super().__init__(
            start_address=address,
            length=1,
            readorwrite=READ_TYPE,
            constant_address=True,
            description=description,
        )

    def process_responses(self, read_responses, write_responses,
                          resolve_future=True):
        e, result = super().process_responses(read_responses, write_responses,
                                              resolve_future=False)
        r = None
        if not e:
            if len(result) != 1:
                e = Exception('Wrong number of results')
            else:
                first_result = result[0]
                if first_result == 1:
                    r = True
                elif first_result == 0:
                    r = False
                else:
                    msg = 'Return value must be 0 to 1 for boolean.  Is {}.'
                    e = Exception(msg.format(result))
        if resolve_future:
            self.resolve_future(e, r)
        return (e, r)


class SetBooleanCommand(AxiCommand):

    def __init__(self, value, address, description=None):
        if value:
            b = 1
        else:
            b = 0
        super().__init__(
            start_address=address,
            data=[b],
            length=1,
            readorwrite=WRITE_TYPE,
            constant_address=True,
            description=description,
        )


class FakeWaitCommand:
    '''
    This is used when we're sending commands to a simulation.
    The `DictCommandHandler` translates it into a bunch of empty
    master-to-slave dictionaries.
    It can be also used when communicating with an FPGA to indicate
    a sleep time.  This is for compatibility of tests.
    '''

    def __init__(self, clock_cycles, sleep_time=0, description=None):
        self.clock_cycles = clock_cycles
        self.axi_commands = []
        self.sleep_time = sleep_time
        self.length = 1
        if description is None:
            self.description = 'Fake wait for {} clock cycles / {} time'.format(
                clock_cycles, sleep_time)
        else:
            self.description = description

    def process_responses(self, read_responses, write_responses,
                          resolve_future=False):
        return None, None

    def get_axi_commands(self):
        return [self]


class CombinedCommand(Command):

    def __init__(self, commands, description):
        self.commands = commands
        super().__init__(description)

    def get_axi_commands(self):
        acs = []
        for command in self.commands:
            acs += command.get_axi_commands()
        return acs

    def process_responses(self, read_responses, write_responses,
                          resolve_future=True):
        first_e = None
        results = []
        for command in self.commands:
            e, result = command.process_responses(
                read_responses, write_responses)
            if (e is not None) and (first_e is None):
                first_e = e
            results.append(result)
        if resolve_future:
            self.resolve_future(first_e, None)
        return first_e, results



import time
import logging
import asyncio

import cocotb
from cocotb import log
from cocotb.triggers import Event, RisingEdge, FallingEdge, Timer

from collections import deque

from axilent import comms

def signal_to_integer(s):
    if not s.value.is_resolvable:
        log.error("Signal is not resolvable")
        import pdb
        pdb.set_trace()
        return None
    else:
        return s.value.integer
    

class CocotbHandler(object):
    '''
    This handler receives `Command` objects and sends their AXI
    commands into a cocotb simulation.
    '''

    def __init__(self, clock_signal, clock_period, axi_signals):
        '''
        `clock_signal`: The signal for the clock.
        `clock_period`: The period of the clock.
        `axi_signals`: A dictionary relating signal names to the cocotb signal objects.
        '''
        self.clock_signal = clock_signal
        self.clock_period = clock_period
        self.signals = axi_signals
        self.write_request_queue = deque()
        self.read_request_queue = deque()
        self.write_response_queue = deque()
        self.read_response_queue = deque()
        cocotb.fork(self.process_write_requests())
        cocotb.fork(self.process_write_responses())
        cocotb.fork(self.process_read_requests())
        cocotb.fork(self.process_read_responses())
        self.command_queue = deque()

    def send(self, command):
        log.info('queuing command {}'.format(command.description))
        self.command_queue.append(command)

    @cocotb.coroutine
    def send_stored_commands(self):
        while self.command_queue:
            command = self.command_queue.popleft()
            log.info('sending command {}'.format(command.description))
            yield self.send_single_command(command)

    @cocotb.coroutine
    def send_single_command(self, command):
        '''
        Sends a Command objects to the FPGA and processes the responses.
        '''
        log.info('really sending command {}'.format(command.description))
        read_futures = []
        write_futures = []
        for ac in command.get_axi_commands():
            log.debug('Command sent for %s.', ac.description)
            if isinstance(ac, comms.FakeWaitCommand):
                log.info('Waiting for {} clock cycles'.format(ac.clock_cycles))
                for dummy_index in range(ac.clock_cycles):
                    yield RisingEdge(self.clock_signal)
                    if dummy_index % 10 == 0:
                        log.info('Got clock cycles {}/{}'.format(dummy_index, ac.clock_cycles))
            else:
                assert(ac.readorwrite in (comms.WRITE_TYPE, comms.READ_TYPE))
                if ac.readorwrite == comms.WRITE_TYPE:
                    for offset, d in enumerate(ac.data):
                        address = ac.start_address
                        if not ac.constant_address:
                            address += offset
                        write_futures.append(self.write(address, d, add_trigger=True))
                else:
                    for index in range(ac.length):
                        address = ac.start_address
                        if not ac.constant_address:
                            address += index
                        read_futures.append(self.read(address, add_trigger=True))
        log.info('Finished sending')

        log.info('{} read futures and {} write_futures'.format(len(read_futures), len(write_futures)))
        for index, future in enumerate(write_futures):
            log.info('Waiting for write future {}'.format(index))
            yield future.trigger.wait()
            log.info('Got write future {}'.format(index))
        for index, future in enumerate(read_futures):
            log.info('Waiting for read future {}'.format(index))
            yield future.trigger.wait()
            log.info('Got read future {}'.format(index))
        write_future_results = [f.result() for f in write_futures]
        write_responses = [comms.AxiResponse(length=1, data=[None], resp=resp)
                           for resp in write_future_results]

        read_future_results = [f.result() for f in read_futures]
        read_responses = [comms.AxiResponse(length=1, data=[data], resp=resp)
                          for resp, data in read_future_results]
        log.info('Got response')
        command.process_responses(read_responses, write_responses)
        log.info('Finished processing')

    def write(self, address, value, add_trigger=True):
        future = asyncio.Future()
        self.write_request_queue.append((address, value, future))
        if add_trigger:
            future.trigger = Event()
        return future

    def read(self, address, add_trigger=True):
        future = asyncio.Future()
        self.read_request_queue.append((address, future))
        if add_trigger:
            future.trigger = Event()
        return future

    @cocotb.coroutine
    def process_write_requests(self):
        while True:
            yield RisingEdge(self.clock_signal)
            yield Timer(int(self.clock_period*0.1))
            if self.write_request_queue:
                address, value, future = self.write_request_queue.popleft()
                self.signals['wvalid'].setimmediatevalue(1)
                self.signals['wdata'].setimmediatevalue(value)
                self.signals['awvalid'].setimmediatevalue(1)
                self.signals['awaddr'].setimmediatevalue(address)
                sent_w = False
                sent_aw = False
                while True:
                    yield FallingEdge(self.clock_signal)
                    yield Timer(int(self.clock_period * 0.4))
                    if not sent_w:
                        sent_w = (self.signals['wready'] == 1)
                    if not sent_aw:
                        sent_aw = (self.signals['awready'] == 1)
                    if sent_w and sent_aw:
                        self.write_response_queue.append(future)
                        break
                    yield RisingEdge(self.clock_signal)
                    yield Timer(int(self.clock_period * 0.1))
                    if sent_aw:
                        self.signals['awvalid'].setimmediatevalue(0)
                    if sent_w:
                        self.signals['wvalid'].setimmediatevalue(0)
            else:
                self.signals['awvalid'].setimmediatevalue(0)
                self.signals['wvalid'].setimmediatevalue(0)

    @cocotb.coroutine
    def process_write_responses(self):
        self.signals['bready'].setimmediatevalue(1)
        while True:
            yield FallingEdge(self.clock_signal)
            yield Timer(int(self.clock_period * 0.4))
            if self.signals['bvalid'] == 1:
                response = self.signals['bresp'].value.integer
                future = self.write_response_queue.popleft()
                future.set_result(response)
                if hasattr(future, 'trigger'):
                    future.trigger.set()

    @cocotb.coroutine
    def process_read_requests(self):
        while True:
            yield RisingEdge(self.clock_signal)
            yield Timer(int(self.clock_period * 0.1))
            if self.read_request_queue:
                address, future = self.read_request_queue.popleft()
                self.signals['arvalid'].setimmediatevalue(1)
                self.signals['araddr'].setimmediatevalue(address)
                sent_ar = False
                while True:
                    yield FallingEdge(self.clock_signal)
                    yield Timer(int(self.clock_period * 0.4))
                    if not sent_ar:
                        sent_ar = (self.signals['arready'].value == 1)
                    if sent_ar:
                        self.read_response_queue.append(future)
                        break
                    yield RisingEdge(self.clock_signal)
                    yield Timer(int(self.clock_period * 0.1))
            else:
                self.signals['arvalid'].setimmediatevalue(0)

    @cocotb.coroutine
    def process_read_responses(self):
        self.signals['rready'].setimmediatevalue(1)
        while True:
            yield FallingEdge(self.clock_signal)
            yield Timer(int(self.clock_period * 0.4))
            if self.signals['rvalid'] == 1:
                response = self.signals['rresp'].value.integer
                data = signal_to_integer(self.signals['rdata'])
                future = self.read_response_queue.popleft()
                future.set_result((response, data))
                if hasattr(future, 'trigger'):
                    future.trigger.set()



import random
import logging

from axilent import handlers, dicts


logger = logging.getLogger(__name__)


class DictAxiTest(object):
    '''
    Wraps a test with prepare and check methods with a test that
    exposes make_input_data and check_output_data methods.
    '''

    def __init__(self, axi_test, terminate_early=False, clock_names=None, reset=True,
                 m2s_name='m2s', s2m_name='s2m'):
        self.axi_test = axi_test
        self.handler = handlers.DictCommandHandler()
        self.terminate_early = terminate_early
        self.clock_names = clock_names
        self.reset = reset
        self.m2s_name = m2s_name
        self.s2m_name = s2m_name

    def make_input_data(self):
        input_data = []
        if self.reset:
            input_data.append({
                'reset': 1,
                self.m2s_name: dicts.make_empty_axi4lite_m2s_dict(),
            })
        self.axi_test.prepare(self.handler)
        m2s = self.handler.make_command_dicts()
        input_data += [{
            'reset': 0,
            self.m2s_name: d,
            } for d in m2s]
        if self.terminate_early:
            input_data = input_data[:random.randint(1, len(input_data))]
        if self.clock_names:
            input_data = {
                self.clock_names[0]: input_data,
                }
            for clock_name in self.clock_names[1:]:
                input_data[clock_name] = []

        return input_data

    def check_output_data(self, input_data, output_data):
        if self.clock_names:
            input_data = input_data[self.clock_names[0]]
            output_data = output_data[self.clock_names[0]]
        if not self.terminate_early:
            ds = [{**ipt[self.m2s_name], **opt[self.s2m_name]} for ipt, opt in zip(
                input_data[1:], output_data[1:])]
            self.handler.consume_response_dicts(ds)
            self.axi_test.check()


class CombinedHandlerTest:

    def __init__(self, subtests, handler):
        self.subtests = subtests
        self.handler = handler

    def make_input_data(self):
        input_data = []
        old_length = 0
        self.lengths = []
        for test in self.subtests:
            input_data += test.make_input_data()
            new_length = len(input_data)
            self.lengths.append((old_length, new_length))
            old_length = new_length
        return input_data

    def check_output_data(self, input_data, output_data):
        # Assume that we ignore the first output_data due to reset.
        reset_index = 0
        response_dicts = [
            {**ipt['config_m2s'], **opt['config_s2m']}
            for ipt, opt in zip(input_data[reset_index+1:], output_data[reset_index+1:])]
        self.handler.consume_response_dicts(response_dicts)
        n_subtests = len(self.subtests)
        for index, indices, test in zip(range(n_subtests), self.lengths, self.subtests):
            logger.debug('Checking output in subtest {}/{}'.format(index+1, n_subtests))
            start_index, end_index = indices
            sub_input_data = input_data[start_index: end_index]
            sub_output_data = output_data[start_index: end_index]
            test.check_output_data(sub_input_data, sub_output_data)



import logging
import random

from slvcodec.test_utils import WrapperTest
from axilent.test_utils import DictAxiTest
from axilent import comms


logger = logging.getLogger(__name__)


class AxiAdderComm(object):
    '''
    Class to communicate with the AxiAdder module.
    '''

    INTA_ADDRESS = 0
    INTB_ADDRESS = 1
    INTC_ADDRESS = 2

    def __init__(self, address_offset, handler):
        '''
        `address_offset` is any addition that is made to the address that is
        consumed during routing.
        `handler` is the object responsible for dispatching the commands.
        '''
        self.handler = handler
        self.address_offset = address_offset
        self.addresses = {
            'intA': address_offset + self.INTA_ADDRESS,
            'intB': address_offset + self.INTB_ADDRESS,
            'intC': address_offset + self.INTC_ADDRESS,
        }

    def add_numbers(self, a, b):
        '''
        A complex complex command that write to two registers and
        then reads from another.
        Sets 'a' and 'b' then reads 'c' (should be a+b)
        '''
        command = AddNumbersCommand(a, b, self.addresses)
        self.handler.send(command)
        return command.future


class AddNumbersCommand(comms.CombinedCommand):
    '''
    A command that writes to the intA and intB registers
    and then reads from the intC register.
    The effect is the add the two inputs.
    '''

    def __init__(self, a, b, addresses):
        write_a_command = comms.SetUnsignedCommand(
            address=addresses['intA'], value=a,
            description='Setting A in AddNumbers',
        )
        write_b_command = comms.SetUnsignedCommand(
            address=addresses['intB'], value=b,
            description='Setting B in AddNumbers',
        )
        read_c_command = comms.GetUnsignedCommand(
            address=addresses['intC'],
            description='Getting C from AddNumbers',
        )
        commands = (write_a_command, write_b_command, read_c_command)
        super().__init__(
            description='Add 2 numbers with AddNumber',
            commands=commands)

    def process_responses(self, read_responses, write_responses, resolve_future=True):
        '''
        Return the third response (from the final read command)
        Don't return any errors.
        '''
        e, result = super().process_responses(read_responses, write_responses, resolve_future=False)
        intc = result[2]
        if resolve_future:
            self.resolve_future(e, intc)
        return e, intc


class AxiAdderTest(object):

    def __init__(self):
        self.expected_intcs = []
        self.intc_futures = []

    def prepare(self, handler):
        '''
        Sends a number of 'add_numbers' commands.
        '''
        comm = AxiAdderComm(address_offset=0, handler=handler)
        n_data = 20
        max_int = pow(2, 16)-1
        logger.debug('preparing data')
        for i in range(n_data):
            inta = random.randint(0, max_int)
            intb = random.randint(0, max_int)
            self.expected_intcs.append(inta + intb)
            future = comm.add_numbers(inta, intb)
            self.intc_futures.append(future)
        # Flush the communication for simulations.
        # Ignored in FPGA.
        handler.send(comms.FakeWaitCommand(clock_cycles=10))

    def check(self):
        '''
        Check that the output of the commands matches the expected values.
        '''
        output_intcs = [f.result() for f in self.intc_futures]
        assert output_intcs == self.expected_intcs
        print('Success!!!!!!!!!!!!!!!')


def make_test(entity, generics, top_params):
    tests = []
    for test_index in range(20):
        terminate_early = random.choice([True, False])
        axi_test = AxiAdderTest()
        tests.append(DictAxiTest(axi_test, terminate_early))
    combined_test = WrapperTest(tests)
    return combined_test


def get_tests():
    test = {
        'core_name': 'axi_adder',
        'entity_name': 'axi_adder',
        'param_sets': [{
            'generic_sets': [{}],
            'top_params': {}
        }],
        'generator': make_test,
        }
    return [test]



from axilent import comms

# Response code for AXI.
OKAY = 0
EXOKAY = 1
SLVERR = 2
DECERR = 3


def make_empty_axi4lite_m2s_dict():
    '''
    Creates and empty master-to-slave AXI dictionary.
    '''
    return {
        'araddr': 0,
        'arprot': 0,
        'arvalid': 0,
        'awaddr': 0,
        'awprot': 0,
        'awvalid': 0,
        'bready': 1,
        'rready': 1,
        'wdata': 0,
        'wstrb': 15,
        'wvalid': 0,
    }


def make_axi4lite_m2s_dict(
        write_address=None, write_data=None, read_address=None,
        bready=0, rready=0):
    m2s = make_empty_axi4lite_m2s_dict()
    if write_address is not None:
        m2s['awaddr'] = write_address
        m2s['awvalid'] = 1
    if write_data is not None:
        m2s['wdata'] = write_data
        m2s['wvalid'] = 1
    if read_address is not None:
        m2s['araddr'] = read_address
        m2s['arvalid'] = 1
    m2s['bready'] = bready
    m2s['rready'] = rready
    return m2s


def make_empty_axi4lite_s2m_dict():
    '''
    Creates and empty slave-to-master AXI dictionary.
    '''
    return {
        'arready': 1,
        'awready': 1,
        'bresp': 0,
        'bvalid': 0,
        'rdata': 0,
        'rresp': 0,
        'rvalid': 0,
        'wready': 1,
    }


def make_axi4lite_s2m_dict(
        read_data=None, arready=0, awready=0, write_response=None, wready=0):
    s2m = make_empty_axi4lite_s2m_dict()
    s2m['arready'] = arready
    s2m['awready'] = awready
    if write_response is not None:
        s2m['bresp'] = write_response
        s2m['bvalid'] = 1
    if read_data is not None:
        s2m['rdata'] = read_data
        s2m['rvalid'] = 1
    s2m['wready'] = wready
    return s2m


def axi_commands_to_axi_dicts(axi_commands):
    ds = []
    for ac in axi_commands:
        if type(ac) in (comms.FakeWaitCommand, ):
            for i in range(ac.clock_cycles):
                ds.append(make_axi4lite_m2s_dict(bready=1, rready=1))
        else:
            for index in range(ac.length):
                d = make_axi4lite_m2s_dict(bready=1, rready=1)
                if ac.constant_address:
                    address = ac.start_address
                else:
                    address = ac.start_address + index
                if ac.readorwrite == comms.READ_TYPE:
                    d['araddr'] = address
                    d['arvalid'] = 1
                else:
                    d['awaddr'] = address
                    d['awvalid'] = 1
                    d['wvalid'] = 1
                    d['wdata'] = ac.data[index]
                    if d['wdata'] < 0:
                        raise Exception(
                            'Command {} is trying to set negative axi data'.format(ac))
                ds.append(d)
    return ds


def axi_dicts_to_axi_responses(axi_dicts):
    assert all([d['bvalid'] is not None for d in axi_dicts])
    assert all([d['rvalid'] is not None for d in axi_dicts])
    # Make sure that no inputs got dropped.
    assert all([d['arready'] for d in axi_dicts if d['arvalid']])
    assert all([d['awready'] for d in axi_dicts if d['awvalid']])
    assert all([d['wready'] for d in axi_dicts if d['wvalid']])
    write_ds = [d for d in axi_dicts if d['bvalid'] and d['bready']]
    read_ds = [d for d in axi_dicts if d['rvalid'] and d['rready']]
    write_responses = [
        comms.AxiResponse(length=1, data=[None], resp=d['bresp']) for d in write_ds]
    read_responses = [
        comms.AxiResponse(length=1, data=[d['rdata']], resp=d['rresp'])
        for d in read_ds]
    return read_responses, write_responses



'''
Python tools for creating and parsing AXI communications.
'''

import logging
import time

from axilent import dicts, comms

logger = logging.getLogger(__name__)


class ConnCommandHandler(object):
    '''
    This handler receives `Command` objects and sends their AXI
    commands over the passed in `Connection`.  The responses are
    then processed by the `Command` object that send them.

    This handler is useful to simplify communication with the FPGA.
    '''

    def __init__(self, conn):
        '''
        `conn`: A `Connection` object that the handler uses to communicate
                with the FPGA.
        '''
        self.conn = conn

    def send(self, command):
        '''
        Sends a Command objects to the FPGA and processes the responses.
        '''
        read_rs = []
        write_rs = []
        for ac in command.get_axi_commands():
            logger.debug('Sending command for %s.', ac.description)
            if isinstance(ac, comms.FakeWaitCommand):
                time.sleep(ac.sleep_time)
            else:
                assert(ac.readorwrite in (comms.WRITE_TYPE, comms.READ_TYPE))
                if ac.readorwrite == comms.WRITE_TYPE:
                    if ac.constant_address:
                        r = self.conn.write_repeat(
                            address=ac.start_address, data=ac.data)
                    else:
                        r = self.conn.write(address=ac.start_address, data=ac.data)
                    write_rs.append(r)
                else:
                    if ac.constant_address:
                        r = self.conn.read_repeat(address=ac.start_address, length=ac.length)
                    else:
                        r = self.conn.read(address=ac.start_address, length=ac.length)
                    read_rs.append(r)
        command.process_responses(read_rs, write_rs)


class DictCommandHandler(object):
    '''
    This handler receives `Command` objects and stores them.
    When the `make_command_dicts` method is called the handler returns
    a list of AXI master-to-slave dictionaries that specify the commands.
    It can also parse the output AXI slave-to-master dictionaries from
    a simulation.

    This handler is useful to fake communication when running simulations.
    '''

    def __init__(self):
        self.unsent_commands = []
        self.sent_commands = []

    def send(self, command):
        assert(not isinstance(command, list))
        self.unsent_commands.append(command)

    def get_axi_commands(self):
        axi_commands = []
        while self.unsent_commands:
            command = self.unsent_commands.pop(0)
            axi_commands += command.get_axi_commands()
            self.sent_commands.append(command)
        return axi_commands

    def process_responses(self, read_responses, write_responses):
        for command in self.sent_commands:
            command.process_responses(read_responses, write_responses)

    def consume_response_dicts(self, ds):
        read_responses, write_responses = dicts.axi_dicts_to_axi_responses(ds)
        self.process_responses(read_responses, write_responses)

    def make_command_dicts(self):
        acs = self.get_axi_commands()
        ads = dicts.axi_commands_to_axi_dicts(acs)
        return ads


class PYNQHandler(object):
    '''
    This handler receives `Command` objects and sends their AXI
    commands over the passed in PYNQ DefaultIP driver.

    This handler is useful to simplify communication with the FPGA.
    '''

    def __init__(self, driver):
        '''
        `driver`: A PYNQ DefaultIP driver with `read` and `write` methods.
                The passed driver has addresses indexing bytes, whereas
                this handler assumes addresses index 32 bits.
        '''
        self.driver = driver

    def send(self, command):
        '''
        Sends a Command objects to the FPGA and processes the responses.
        '''
        read_rs = []
        write_rs = []
        for ac in command.get_axi_commands():
            logger.debug('Command sent for %s.', ac.description)
            if isinstance(ac, comms.FakeWaitCommand):
                time.sleep(ac.sleep_time)
            else:
                assert(ac.readorwrite in (comms.WRITE_TYPE, comms.READ_TYPE))
                if ac.readorwrite == comms.WRITE_TYPE:
                    if ac.constant_address:
                        for d in ac.data:
                            self.driver.write(ac.start_address*4, d)
                            write_rs.append(comms.AxiResponse(length=1, data=[None], resp=0))
                    else:
                        for offset, d in enumerate(ac.data):
                            self.driver.write((ac.start_address+offset)*4, d)
                            write_rs.append(comms.AxiResponse(length=1, data=[None], resp=0))
                else:
                    if ac.constant_address:
                        for index in range(ac.length):
                            response = self.driver.read(ac.start_address*4)
                            read_rs.append(comms.AxiResponse(length=1, data=[response], resp=0))
                    else:
                        for index in range(ac.length):
                            response = self.driver.read((ac.start_address+index)*4)
                            read_rs.append(comms.AxiResponse(length=1, data=[response], resp=0))
        command.process_responses(read_rs, write_rs)



import os
from setuptools import setup

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='axilent',
    packages=['axilent', 'axilent.examples'],
    package_data={'': ['cores/*.core', 'vhdl/*.vhd']},
    use_scm_version={
        'relative_to': __file__,
        'write_to': 'axilent/version.py',
    },
    setup_requires=['setuptools_scm'],
    author='Ben Reynwar',
    author_email='ben@reynwar.net',
    description=('Tools for describing a sequence of Axi4Lite command'),
    long_description=long_description,
    long_description_content_type='text/x-rst',
    license='MIT',
    keywords=['VHDL', 'hdl', 'rtl', 'FPGA', 'ASIC', 'Xilinx', 'Altera'],
    url='https://github.com/benreynwar/axilent',
    install_requires=[
    ],
    dependency_links=[
    ],
)



import os

from slvcodec import config as slvcodec_config, test_utils
from axilent.examples import axi_adder
from axilent import coresdir

testoutput_dir = os.path.join(os.path.dirname(__file__), '..', 'test_outputs')

def test_axi_adder():
    tests = axi_adder.get_tests()
    slvcodec_config.setup_fusesoc(cores_roots=[coresdir])
    vu = slvcodec_config.setup_vunit(argv=['--dont-catch-exceptions'])
    for coretest in tests:
        test_utils.register_coretest_with_vunit(vu, coretest, testoutput_dir)
    all_ok = vu._main(post_run=None)
    assert all_ok

if __name__ == '__main__':
    test_axi_adder()
